/****************************************************************************
	Bagotronix
	1019 Crossing Brook Way
	Tallahassee, FL 32311

	(C) Copyright 1996 Bagotronix
	All rights reserved

	Filename:  		csfclcd.c
	Programmer(s):	Ivan Baggett
	Description:	Sunflower C(tm) library
					Chickadee(tm) LCD functions
	comments:		Makes use of the PDREM remote console.  In order for this
					to work PDRT186 must be in dynamic mode and the remote
					console window must be the active window.  CONSOLE.OBJ
					must also be linked in.

					All LCD functions herein assume the use of an LCD module
					based on the Hitachi HD44780 display control LSI.
****************************************************************************/


#pragma inline


#include "stdinc.h"
#include "chickade.h"
#include "csfclcd.h"


static COORD pos;						/* current cursor position */
static const COORD LCDmaxpos[] = {
	{	16, 1	},
	{	16, 2	},
	{	20, 1	},
	{	20, 2	},
	{	40, 2	},
	{	40, 4	},
	{	20, 4	}
};
UBYTE LCDmode;					/* on/off, cursor on/off, (no)blink */

static const UBYTE LCD16x1[] = {
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7
};

static const UBYTE LCD16x2[] = {
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
					0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
					0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf
};

static const UBYTE LCD20x1[] = {
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
					0x90, 0x91, 0x92, 0x93
};

static const UBYTE LCD20x2[] = {
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
					0x90, 0x91, 0x92, 0x93,
					0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
					0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
					0xd0, 0xd1, 0xd2, 0xd3
};

static const UBYTE LCD20x4[] = {
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
					0x90, 0x91, 0x92, 0x93,
					0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
					0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
					0xd0, 0xd1, 0xd2, 0xd3,
					0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0x9b,
					0x9c, 0x9d, 0x9e, 0x9f, 0xa0, 0xa1, 0xa2, 0xa3,
					0xa4, 0xa5, 0xa6, 0xa7,
					0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xdb,
					0xdc, 0xdd, 0xde, 0xdf, 0xe0, 0xe1, 0xe2, 0xe3,
					0xe4, 0xe5, 0xe6, 0xe7
};

/*	!!! NOTE:  DO NOT TRY TO USE 40X2 TWICE FOR 40X4. COMPILER MERGES
		DUPLICATE SEQUENCES AND CREATES BUG !!! */
static const UBYTE LCD40x2[] = {
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
					0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
					0x98, 0x99, 0x9a, 0x9b,	0x9c, 0x9d, 0x9e, 0x9f,
					0xa0, 0xa1, 0xa2, 0xa3,	0xa4, 0xa5, 0xa6, 0xa7,
					0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
					0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
					0xd0, 0xd1, 0xd2, 0xd3,	0xd4, 0xd5, 0xd6, 0xd7,
					0xd8, 0xd9, 0xda, 0xdb,	0xdc, 0xdd, 0xde, 0xdf,
					0xe0, 0xe1, 0xe2, 0xe3,	0xe4, 0xe5, 0xe6, 0xe7
};

static const UBYTE LCD40x4[] = {	/* 40x2 pattern is duplicated */
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
					0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
					0x98, 0x99, 0x9a, 0x9b,	0x9c, 0x9d, 0x9e, 0x9f,
					0xa0, 0xa1, 0xa2, 0xa3,	0xa4, 0xa5, 0xa6, 0xa7,
					0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
					0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
					0xd0, 0xd1, 0xd2, 0xd3,	0xd4, 0xd5, 0xd6, 0xd7,
					0xd8, 0xd9, 0xda, 0xdb,	0xdc, 0xdd, 0xde, 0xdf,
					0xe0, 0xe1, 0xe2, 0xe3,	0xe4, 0xe5, 0xe6, 0xe7,
					0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
					0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
					0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,
					0x98, 0x99, 0x9a, 0x9b,	0x9c, 0x9d, 0x9e, 0x9f,
					0xa0, 0xa1, 0xa2, 0xa3,	0xa4, 0xa5, 0xa6, 0xa7,
					0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,
					0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
					0xd0, 0xd1, 0xd2, 0xd3,	0xd4, 0xd5, 0xd6, 0xd7,
					0xd8, 0xd9, 0xda, 0xdb,	0xdc, 0xdd, 0xde, 0xdf,
					0xe0, 0xe1, 0xe2, 0xe3,	0xe4, 0xe5, 0xe6, 0xe7
};

static const UBYTE *LCDMMxY[] = {
	LCD16x1,	/* type 0 16x1 */
	LCD16x2,	/* type 1 16x2 */
	LCD20x1,	/* type 2 20x1 */
	LCD20x2,	/* type 3 20x2 */
	LCD40x2,	/* type 4 40x2 */
	LCD40x4,	/* type 5 40x4 (40X2 pattern is duplicated) */
	LCD20x4		/* type 6 20x4 */
};

static UBYTE LCDtype = 0;	/* default LCD type */

static UBYTE cval;
#if 1		// rev 7/20/96
static const UWORD contrast[24][2] = {
	{	12, 20	},
	{	13, 20	},
	{	14, 20	},
	{	15, 20	},
	{	16, 20	},
	{	17, 20	},
	{	18, 20	},
	{	19, 20	},

	{	20, 20	},
	{	20, 19	},
	{	20, 18	},
	{	20, 17	},
	{	20, 16	},
	{	20, 15	},
	{	20, 14	},
	{	20, 13	},
	{	20, 12	},
	{	20, 11	},
	{	20, 10	},
	{	20, 9 	},
	{	20, 8 	},
	{	20, 7 	},
	{	20, 6 	},
	{	20, 5 	}
};
#endif

#if 0
// these contrast constants chosen to give fixed interval of 54925 when added
// formulas: cnta = ((Vee - 5) / -10) * 54925, cntb = 54925 - cnta
//		where Vee is the LCD contrast voltage in Volts
//		NOTE:  use of these constants requires C19 = 47uF !!!
//				use other set of constants otherwise
static const UWORD contrast[51][2] = {
	{       15928, 38997  },	// 2.10 V
	{       16697, 38228  },	// 1.96 V
	{       17466, 37459  },	// 1.82 V
	{       18235, 36690  },	// 1.68 V
	{       19004, 35921  },	// 1.54 V
	{       19773, 35152  },	// 1.40 V
	{       20542, 34383  },	// 1.26 V
	{       21311, 33614  },	// 1.12 V
	{       22080, 32845  },	// 0.98 V
	{       22849, 32076  },	// 0.84 V
	{       23618, 31307  },	// 0.70 V
	{       24387, 30538  },	// 0.56 V
	{       25156, 29769  },	// 0.42 V
	{       25925, 29000  },	// 0.28 V
	{       26694, 28231  },	// 0.14 V

	{       27463, 27462  },	// 0 V
	{       28231, 26694  },    // -0.14 V
	{       29000, 25925  },	// -0.28 V
	{       29769, 25156  },	// -0.42 V
	{       30538, 24387  },	// -0.56 V
	{       31307, 23618  },	// -0.70 V
	{       32076, 22849  },	// -0.84 V
	{       32845, 22080  },	// -0.98 V
	{       33614, 21311  },	// -1.12 V
	{       34383, 20542  },	// -1.26 V
	{       35152, 19773  },	// -1.40 V
	{       35921, 19004  },	// -1.54 V
	{       36690, 18235  },	// -1.68 V
	{       37459, 17466  },	// -1.82 V
	{       38228, 16697  },	// -1.96 V
	{       38997, 15928  },	// -2.10 V
	{       39766, 15159  },    // -2.24 V
	{       40535, 14390  },	// -2.38 V
	{       41304, 13621  },	// -2.52 V
	{       42073, 12852  },	// -2.66 V
	{       42842, 12083  },	// -2.80 V
	{       43610, 11315  },	// -2.94 V
	{       44379, 10546  },	// -3.08 V
	{       45148,  9777  },	// -3.22 V
	{       45917,  9008  },	// -3.36 V
	{       46686,  8239  },	// -3.50 V
	{       47455,  7470  },	// -3.64 V
	{       48224,  6701  },	// -3.78 V
	{       48993,  5932  },	// -3.92 V
	{       49762,  5163  },	// -4.06 V
	{       50531,  4394  },	// -4.20 V
	{       51300,  3625  },	// -4.34 V
	{       52069,  2856  },	// -4.48 V
	{       52838,  2087  },	// -4.62 V
	{       53607,  1318  },	// -4.76 V
	{       54376,   549  }		// -4.90 V
};
#endif

/****************************************************************************
LCDsetcontrast():	sets the LCD contrast to the given value

	arguments:	the contrast value to use (a dimensionless number)
	returns:  nothing
		resources used:  timer 0
		comments:  timer 2 can be used as prescaler to timer 0.  This is necessary
				in systems using BIOS, because timer 0 is also used as the
				BIOS tick.  For 8 MHz CPU, timer 2 should divide by 2, and
				timer 0 (max count A + max count B) should be 54925.  This
				gives 18.206507365 Hz BIOS tick, just like PC.
	verified:  7/20/96
			   3/16/97
****************************************************************************/
void LCDsetcontrast (UBYTE c) {
	if (c < sizeof contrast / sizeof (UWORD) / 2) {
		cval = c;
		while (inport (T0CON) & (0x1000));  // wait for max count B to be free
		outport(T0CMPB , contrast[cval][1]); // program max count B
		while (!(inport (T0CON) & (0x1000)));  // wait for max count A to be free
		outport(T0CMPA , contrast[cval][0]); // program max count A
#if 0           // rev 7/20/96
		/* stop counting */
		outport (T0CON, (inport (T0CON) & (~0x8000)) | 0x4000);
		/* Clear Timer 0 Count Register to avoid occasional display flash */
		outport(T0CNT , 0);
		/* Timer Compare Register A */
		outport(T0CMPA , contrast[cval][0]);
		/* Timer Compare Register B */
		outport(T0CMPB , contrast[cval][1]);
		/* restart counting */
		outport (T0CON, inport (T0CON) | 0xc000);
#endif
	}
}


/****************************************************************************
LCDgetcontrast():	gets the current LCD contrast value

	arguments:	none
	returns:  	the contrast value in use (a dimensionless number)
	resources used:	none
	comments:
	verified:  7/20/96
****************************************************************************/
UBYTE LCDgetcontrast (void) {
	return cval;
}


/****************************************************************************
LCDdarken():	sets the LCD contrast to the next darker value

	arguments:	none
	returns:  nothing
	resources used:  timer 0
	comments:
	verified:  7/20/96
****************************************************************************/
void LCDdarken (void) {
	UBYTE c;

	c = LCDgetcontrast();
	++c;
	LCDsetcontrast (c);
}


/****************************************************************************
LCDlighten():	sets the LCD contrast to the next lighter value

	arguments:	none
	returns:  nothing
	resources used:  timer 0
	comments:
	verified:  7/20/96
****************************************************************************/
void LCDlighten (void) {
	UBYTE c;

	c = LCDgetcontrast();
	--c;
	LCDsetcontrast (c);

}


/****************************************************************************
LCDinit():	turns on LCD and initializes it to the default mode

	arguments:	the LCD type
	returns:  nothing
	resources used:  LCD, 16C452 parallel port, timer 0
	comments:
	verified:	7/20/96
****************************************************************************/
void LCDinit (UBYTE type) {
	LCDtype = type;
	LCDmode = LCD_DISPLAY_ON | LCD_CURSOR_ON;	/* default mode */
	/*	Wait at least 15 ms because the CPU may have gotten here before then,
		which would not give the LCD enough time to power up.  Optrex manual
		says wait >= 15 ms after power-up for 4-bit interface.
	*/
	CHKdelay (DLY (15000000));
	LCDwrite (0x28, COMMAND, 0);         /* 4 bit */
	CHKdelay (DLY (4100000));                   /* delay 4.1 ms */
	LCDwrite (0x28, COMMAND, 0);
	CHKdelay (DLY (4100000));                   /* delay 4.1 ms */
	LCDwrite (0x28, COMMAND, 0);
	CHKdelay (DLY (4100000));                   /* delay 4.1 ms */
	LCDwrite (0x0f, COMMAND, 0);     /* on, cursor on, blink */
	LCDwrite (0x06, COMMAND, 0);     /* shift cursor, increment */
	LCDwrite (0x01, COMMAND, 0);     /* clear display */
	/* if 4x40, do 2nd controller */
	if (LCDtype == LCD40X4) {
		LCDwrite (0x28, COMMAND, 1);     /* 4 bit */
		CHKdelay (DLY (4100000));                   /* delay 4.1 ms */
		LCDwrite (0x28, COMMAND, 1);
		CHKdelay (DLY (4100000));                   /* delay 4.1 ms */
		LCDwrite (0x28, COMMAND, 1);
		CHKdelay (DLY (4100000));                   /* delay 4.1 ms */
		LCDwrite (0x0f, COMMAND, 1);     /* on, cursor on, blink */
		LCDwrite (0x06, COMMAND, 1);     /* shift cursor, increment */
		LCDwrite (0x01, COMMAND, 1);     /* clear display */
	}

	/* Timer 0 control register
	*   Enable = (1) Enabled
	*   Inhibit = (1) Enable Change Allowed
	*   Interrupt = (0) No Interrupt
	*   Retrigger = (0) Disabled
	*   Prescaler = (0) Disabled
	*   External Clock = (0) Internal Clock
	*   Alternate Compare Register = (1) Enabled
	*   Continuous Mode = (1) Enabled
	*/
	outport(T0CON , 0xC003);
	LCDsetcontrast (8);		/* set contrast to default (about 0 V) */
	/* initialize screen position */
	LCDgotoxy (1, 1);
}


/***************************************************************************
LCDwrite:  write data or command to LCD

	arguments:  char of data or command, and type (COMMAND or DATA), which
		controller (if 4x40)
	returns:  nothing
	comments:
	verified:  7/21/96
**********************************/
void LCDwrite (UBYTE c, UBYTE ctype, UBYTE which) {
asm {
	mov     dx, LPTDATA
	in      al, dx
	}
	if (ctype == COMMAND) {
asm     and     al, NOT 0x02;       /* PD1 = 0 for command */
	}
	else {
asm     or      al, 0x02;           /* PD1 = 1 for data */
	}

asm {
	out     dx, al

	/* give upper nibble to LCD */
	mov     bl, c
	shr     bl, 1
	shr     bl, 1
	and     al, NOT 0x3c
	and     bl, 0x3c            // chop off 2 msb's and 2 lsb's
	or      al, bl
	out     dx, al

	/* give E signal to correct controller if using 4x40 LCD module */
	cmp     byte ptr which, 0
	jne     LCDwrite_1           // which == 1 -> 2nd controller

	or      al, 0x40            // E1 = 1
	out     dx, al
	and     al, NOT 0x40        // E1 = 0
	jmp     LCDwrite_2
	}
LCDwrite_1:
asm {
	or      al, 0x80            // E2 = 1
	out     dx, al
	and     al, NOT 0x80        // E2 = 0
	}
LCDwrite_2:
asm {
	out     dx, al

	/* give lower nibble to LCD */
	mov     bl, c
	shl     bl, 1
	shl     bl, 1
	and     al, NOT 0x3c
	and     bl, 0x3c            // chop off 2 msb's and 2 lsb's
	or      al, bl
	out     dx, al

	/* give E signal to correct controller if using 4x40 LCD module */
	cmp     byte ptr which, 0
	jne     LCDwrite_3           // which == 1 -> 2nd controller

	or      al, 0x40            // E1 = 1
	out     dx, al
	and     al, NOT 0x40        // E1 = 0
	jmp     LCDwrite_4
	}
LCDwrite_3:
asm {
	or      al, 0x80            // E2 = 1
	out     dx, al
	and     al, NOT 0x80        // E2 = 0
	}
LCDwrite_4:
asm {
	out     dx, al
	}

	/* 40 us delay (1640 us delay for Clr Disp and Home commands) */
	if (c <= 3 && ctype == COMMAND) {
		/* delay at least 1640 us */
		CHKdelay (DLY (1640000));
	}
	else {
		/* delay 40 us */
		CHKdelay (DLY (40000));
	}
}


/****************************************************************************
LCDputc():  puts char at current cursor position, advances cursor

	arguments:  the char
	returns:  nothing
	resources used:  LCD, 16C452 parallel port
	comments:	like the C library putch() function
	verified:	7/20/96
****************************************************************************/
int LCDputc (int c) {
	LCDwrite (c, DATA, (LCDtype == LCD40X4 && pos.y > 2) ? 1 : 0);
	++pos.x;
	if (pos.x > LCDmaxpos[LCDtype].x) {
		pos.x = 1;
		++pos.y;				/* line wrap */
		if (pos.y > LCDmaxpos[LCDtype].y)
			pos.y = 1;		/* screen wrap */
		LCDgotoxy (pos.x, pos.y);		/* put LCD cursor where it should be */
	}
	return c;
}


/****************************************************************************
LCDin():  determines if the given coordinates are within the LCD display area

	arguments:  the x and y coordinates
	returns:  0 = out of range, non-zero = in range
	resources used:  none
	comments:
	verified:	7/21/96
****************************************************************************/
BOOLEAN LCDin (int x, int y) {
	if (x >= 1 && x <= (UWORD)LCDmaxpos[LCDtype].x
		&& y >= 1 && y <= (UWORD)LCDmaxpos[LCDtype].y)
		return TRUE;
	else
		return FALSE;
}


/****************************************************************************
LCDgotoxy():  sets the LCD's cursor to a given position

	arguments:  the x and y coordinates
	returns:  nothing
	resources used:  LCD, 16C452 parallel port
	comments:  like the C library gotoxy() function
	verified:  7/20/96
****************************************************************************/
void LCDgotoxy (int x, int y) {

	if (LCDin (x, y)) {
		/* set DDRAM address */
		LCDwrite (*(LCDMMxY[LCDtype] + (y - 1) * (UWORD)LCDmaxpos[LCDtype].x + x - 1),
			COMMAND, (LCDtype == LCD40X4 && y > 2) ? 1 : 0);
		pos.x = (UBYTE)x;
		pos.y = (UBYTE)y;
		LCDsetmode (LCDgetmode());	/* turn the proper cursor to the proper mode */
	}
}


/****************************************************************************
LCDclreol():  clear from present position to end of current line

	arguments:  none
	returns:  nothing
	resources used:  LCD, 16C452 parallel port
	comments:	like C library clreol() function
	verified:	7/21/96
****************************************************************************/
void LCDclreol (void) {
	UBYTE i;
	UBYTE x = pos.x;		/* save present position */
	UBYTE y = pos.y;

	for (i = pos.x; i <= LCDmaxpos[LCDtype].x; i++)
		LCDputc (' ');	/* put spaces */
	LCDgotoxy (x, y);	/* restore position */
}


/****************************************************************************
LCDclrscr():  clears entire LCD screen

	arguments:  none
	returns:  nothing
	resources used:  LCD, 16C452 parallel port
	comments:  like the C library clrscr() function
	verified:	7/21/96
****************************************************************************/
void LCDclrscr (void) {
	LCDwrite (0x01, COMMAND, 0);
	if (LCDtype == LCD40X4)
		LCDwrite (0x01, COMMAND, 1);
	LCDgotoxy (1, 1);
}


/****************************************************************************
LCDsetmode():  sets the LCD's operating mode

	arguments:  the mode
	returns:  nothing
	resources used:  LCD, 16C452 parallel port
	comments:  controls LCD on/off, cursor on/off, blink/no blink,
		nonzeros for LCD on, cursor on, blink
	verified:	7/21/96
****************************************************************************/
void LCDsetmode (UBYTE mode) {
	UBYTE c = 0x08 | (mode & 0x07);

	LCDmode = c;
	if (LCDtype != LCD40X4)
		LCDwrite (c, COMMAND, 0);
	else {
		if (pos.y > 2) {
			LCDwrite (0x08 | LCD_DISPLAY_ON, COMMAND, 0);	/* turn 1st LSI off */
			LCDwrite (c, COMMAND, 1);
		}
		else {
			LCDwrite (c, COMMAND, 0);
			LCDwrite (0x08 | LCD_DISPLAY_ON, COMMAND, 1);	/* turn 2nd LSI off */
		}
	}
}


/****************************************************************************
LCDgetmode():  gets the LCD's operating mode

	arguments:  none
	returns:  the LCD operating mode
	resources used:  none
	comments:  reports the LCD on/off, cursor on/off, blink/no blink status
	verified:	7/21/96
****************************************************************************/
UBYTE LCDgetmode (void) {
	return LCDmode;
}


/****************************************************************************
LCDputs():  puts string beginning at current cursor position

	arguments:  ptr to the string
	returns:  1
	resources used:  LCD, 16C452 parallel port
	comments:	like the C library puts() function
		Does not stop until null terminator is encountered.  This means you
		must ensure the string does not extend past the end of the display
		line if you do not want it to wrap to the next line or the top of the
		display.
	verified:	7/22/96
****************************************************************************/
int LCDputs (const char *s) {
	char *t;

	t = (char *) s;							/* copy pointer */
	while (*t)
		LCDputc (*t++);
	return 1;
}


/****************************************************************************
LCDwherex():  returns x position of cursor

	arguments:  none
	returns:  x coordinate of cursor
	resources used:  none
	comments:	like the C library wherex() function
	verified:	7/22/96
****************************************************************************/
int LCDwherex (void) {
	return ((UWORD)pos.x);
}


/****************************************************************************
LCDwherey():  returns y position of cursor

	arguments:  none
	returns:  the y coordinate of cursor
	resources used:  none
	comments:  like the C library wherey() functions
	verified:	7/22/96
****************************************************************************/
int LCDwherey (void) {
	return ((UWORD)pos.y);
}


/****************************************************************************
LCDmaxx():  returns maximum x position of cursor

	arguments:  none
	returns:  the maximum x coordinate of cursor
	resources used:  none
	comments:
	verified:	7/22/96
****************************************************************************/
int LCDmaxx (void) {
	return (UWORD)LCDmaxpos[LCDtype].x;
}


/****************************************************************************
LCDmaxy():  returns maximum y position of cursor

	arguments:  none
	returns:  the maximum y coordinate of cursor
	resources used:  none
	comments:
	verified:	7/22/96
****************************************************************************/
int LCDmaxy (void) {
	return (UWORD)LCDmaxpos[LCDtype].y;
}


/****************************************************************************
LCD_CGRAMset():  programs LCD CG RAM with a bit pattern for special chars

	arguments:  CG RAM address (0-7), ptr to bit pattern array
	returns:	nothing
	resources used:  LCD, 16C452 parallel port
	comments:  due to null terminator string convention, you cannot embed a
		reference to ASCII value 0 in a string, therefore, to use special
		character 0, you must 'putc()' it.
	verified:	7/22/96
				1/31/97 modified CG RAM address test to allow char table
					positions 0-7 to be written.  These are echoed in
					positions 8-15.
****************************************************************************/
void LCD_CGRAMset (UBYTE d, const UBYTE *s) {
	int i;
	COORD pos;

/*	if (d < 2) {	7/22/96 version */
	if (d < 8) {
		pos.x = LCDwherex();	/* save current position */
		pos.y = LCDwherey();
		d *= 8;
		/* write CG RAM */
		LCDgotoxy (1, 1);
		for (i = 0; i < 8; i++) {
			LCDwrite ((d + i) | 0x40, COMMAND, 0);	/* set CG RAM address */
			LCDwrite (*(s + i), DATA, 0);		/* write pattern to CG RAM */
		}
		LCDgotoxy (1, 1);						/* set DD RAM address */
		if (LCDtype == LCD40X4) {				/* do 2nd controller */
			LCDgotoxy (1, 3);
			for (i = 0; i < 8; i++) {
				LCDwrite ((d + i) | 0x40, COMMAND, 1);	/* set CG RAM address */
				LCDwrite (*(s + i), DATA, 1);		/* write pattern to CG RAM */
			}
			LCDgotoxy (1, 3);					/* set DD RAM address */
		}
		LCDgotoxy (pos.x, pos.y);			/* restore original position */
	}
}
